/*
 * file EditRecord.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.stp.client.samples.EditRecord
 *
 * (C) Copyright IBM Corporation 2005, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.stp.client.samples;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Hashtable;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;
import javax.wvcm.Folder;
import javax.wvcm.PropertyRequestItem;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.PropertyRequestItem.PropertyRequest;

import com.ibm.rational.wvcm.stp.StpException;
import com.ibm.rational.wvcm.stp.StpLocation;
import com.ibm.rational.wvcm.stp.StpProperty;
import com.ibm.rational.wvcm.stp.StpProperty.MetaPropertyName;
import com.ibm.rational.wvcm.stp.cq.CqAction;
import com.ibm.rational.wvcm.stp.cq.CqAttachment;
import com.ibm.rational.wvcm.stp.cq.CqFieldDefinition;
import com.ibm.rational.wvcm.stp.cq.CqFieldValue;
import com.ibm.rational.wvcm.stp.cq.CqProvider;
import com.ibm.rational.wvcm.stp.cq.CqRecord;

/**
 * A sample CM API application that allows the user to browse to a record by
 * executing a query and selecting a record from the result set. Within the
 * record display the fields of the record can be modified under an applicable
 * action.
 */
public class EditRecord {
    /**
     * An extension of the ViewRecord.Viewer that adds the ability to start an
     * action under which fields of the record can be modified.
     */
    static class Viewer extends ViewRecord.Viewer {
        Viewer(CqProvider provider) { super(provider); }
        
        /**
         * Reads the legal actions for the current state of a given record
         * then displays the fields of the record 
         * (using {@link ViewRecord.Viewer#showRecord}) with a drop-down 
         * containing the actions and buttons the user can use to initiate an
         * edit or add or remove an attachment.
         * @see com.ibm.rational.stp.client.samples.ExecuteQuery.Viewer#view(com.ibm.rational.wvcm.stp.cq.CqRecord)
         */
        public JFrame view(CqRecord selected)
        {
            try {
                final CqRecord record =
                        (CqRecord)selected.doReadProperties(RECORD_PROPERTIES);
                final JButton start = new JButton("Edit");
                final JComboBox choices = new JComboBox();
        
                for (CqAction a: record.getLegalActions()) {
                    if (a.getType() != CqAction.Type.IMPORT
                        && a.getType() != CqAction.Type.SUBMIT
                        && a.getType() != CqAction.Type.BASE
                        && a.getType() != CqAction.Type.RECORD_SCRIPT_ALIAS)
                        choices.addItem(a);
                }
                
                final JButton add = new JButton("Attach");
                final JButton remove = new JButton("Detach");
                final ViewRecord.RecordFrame frame = 
                    showRecord("View: ", record, choices.getItemCount()==0? null: 
                                 new JComponent[]{choices, start, add, remove});
                
                add.addActionListener(new ActionListener(){
                    /**
                     * Prompts the user for the name of a file to be attached to
                     * the selected field. If so provided, the content of the 
                     * file is read into the ClearQuest database as an attachment.
                     */
                    public void actionPerformed(ActionEvent arg0)
                    {
                        if (!ViewRecord.isAttachmentList(frame.m_fields, 
                                                         frame.m_table.getSelectedRow()))
                            return;
        
                        JFileChooser chooser = new JFileChooser();
        
                        if (chooser.showOpenDialog(frame) == 
                                JFileChooser.APPROVE_OPTION) {
                            try {
                               String filename = 
                                   chooser.getSelectedFile().getAbsolutePath();
                               CqFieldValue field = 
                                   frame.m_fields.get(frame.m_table.getSelectedRow());
                               StpLocation aLoc = (StpLocation) 
                                   ((Folder)field.getValue()).location()
                                       .child("new" + System.currentTimeMillis());
                               CqAttachment attachment = 
                                   record.cqProvider().cqAttachment(aLoc);
                               
                               attachment = attachment
                                   .doCreateAttachment(filename, 
                                                       null, 
                                                       CqProvider.DELIVER_ALL);
                               
                               JOptionPane.showMessageDialog(frame, 
                                  "Added '" + filename + "' as " + attachment);
                               
                               frame.dispose();
                               view(record);
                            } catch(Throwable t) 
                                { Utilities.exception(frame, "Add Attachment", t);}
                        }}});
                
                remove.addActionListener(new ActionListener(){
                    /**
                     * Allows the user to select an attachment associated with
                     * the selected field. If the user selects such an attachment
                     * it is deleted.
                     */
                    public void actionPerformed(ActionEvent arg0)
                    {
                        if (!ViewRecord.isAttachmentList(frame.m_fields, frame.m_table.getSelectedRow()))
                            return;
                        
                        try {
                           CqAttachment attachment = ViewRecord.selectAttachment
                               (frame, frame.m_fields, frame.m_table.getSelectedRow(), "Remove");
                           
                           if (attachment != null) {
                               attachment.doUnbindAll(null);
                               frame.dispose();
                               view(record);
                           }
                        } catch(Throwable t) 
                            { Utilities.exception(frame, "Remove Attachment", t);}
                        }
                    }
                );
        
                start.addActionListener(new ActionListener(){
                    /**
                     * Starts editing of the record using the action currently
                     * selected by the combo box on the view dialog.
                     */
                    public void actionPerformed(ActionEvent arg0)
                    {
                        CqAction action = (CqAction)choices.getSelectedItem();
                        
                        try {
                            if (action.getType() == CqAction.Type.DUPLICATE) {
                                String id = JOptionPane.showInputDialog
                                    (frame, "Enter ID of duplicated record");
                                
                                if (id == null) return;
                                
                                action.argumentMap(new Hashtable<String, Object>());
                                action.argumentMap()
                                    .put("original", 
                                         record.cqProvider()
                                             .cqRecord((StpLocation)record
                                             .location().parent().child(id)));
                            }
        
                            edit(record, action);
                            frame.dispose();
                        } catch (Exception ex) {
                            Utilities.exception(frame, "Duplicate Action", ex);
                        }
                    }
                });
                
                return frame;
            } catch (WvcmException ex){
                ex.printStackTrace();
            }
            
            return null;
        }

        /**
         * Constructs a new change context in which the given record can be
         * edited, initiates the editing using the given action, and then
         * displays the editable record in a new window for editing.
         * @param selected A Record proxy for the record to be edited. Must not
         * be null but needs no properties defined.
         * @param action An Action proxy for the action to be used for editing.
         * Must not be null. 
         */
        public void edit(CqRecord selected, CqAction action)
        {
            try {
                CqProvider context = m_provider;
                CqRecord record = context.cqRecord(selected.stpLocation());
        
                record =
                    (CqRecord) record.setAction(action)
                        .doWriteProperties(RECORD_PROPERTIES,
                                           CqProvider.HOLD);
                editRecord("Edit: ", record, selected);
            } catch (WvcmException ex){
                Utilities.exception(null, "Start Action", ex);
            }
        }
        
        /**
         * Displays the fields of a record and allows the user to modify those
         * fields (as permitted by the schema) until the changes are either
         * committed back to the database or abandoned altogether by the user.
         * @param title The title string for the window in which the record is
         * displayed.
         * @param record The Record proxy for the editable version of the record.
         * @param selected The Record proxy for the original version of the
         * record. Used to redisplay the record (in its original change context)
         * after the editable version has either been committed or abandoned.
         * @throws WvcmException If the properties required for the editable
         * record are not defined by the editable record proxy. 
         */
        JFrame editRecord(
            final String title, 
            final CqRecord record, 
            final CqRecord selected
            ) throws WvcmException 
       {
            final JFrame frame = 
                new JFrame(title + record.getUserFriendlyLocation().toString());
            JPanel panel = new JPanel(new BorderLayout());
            JPanel buttons = new JPanel(new FlowLayout());
            final JButton show = new JButton("View");
            final JButton add = new JButton("Add");
            final JButton remove = new JButton("Remove");
            final JButton cancel = new JButton("Cancel");
            final StpProperty.List<CqFieldValue<?>> fields = 
                record.getAllFieldValues();
            TableModel dataModel = new AbstractTableModel() {
                public int getColumnCount() { return fieldMetaProperties.length; }
                public int getRowCount() { return fields.size();}
                public String getColumnName(int col)
                { return fieldMetaProperties[col].getRoot().getName(); }
                public Object getValueAt(int row, int col) 
                    { 
                        try {
                            return fields.get(row)
                                .getMetaProperty((MetaPropertyName<?>)
                                                 fieldMetaProperties[col].getRoot());
                        } catch(Throwable ex) {
                            if (ex instanceof StpException) {
                                return ((StpException)ex).getStpReasonCode();  
                              } else {
                                  String name = ex.getClass().getName();
                                  return name.substring(name.lastIndexOf(".")+1);
                              }
                        }
                    }
                
                /**
                 * Determines whether or not the specified cell should be 
                 * editable. A cell is editable only if it corresponds to the
                 * VALUE meta-property of a field whose REQUIREDNESS is not
                 * READ_ONLY
                 */ 
                public boolean isCellEditable(int row, int col) {
                    if (fieldMetaProperties[col].getRoot()
                                    .equals(StpProperty.VALUE)) {
                        CqFieldValue field = (CqFieldValue)fields.get(row);
                        
                        try {
                            return field.getRequiredness() 
                                    != CqFieldDefinition.Requiredness.READ_ONLY;
                        } catch (WvcmException ex) {
                            Utilities.exception(frame, "Field Requiredness", ex);
                        }
                    }
                    
                    return false;
                }
                
                /**
                 * Sets a new value into the record field that corresponds to
                 * the specified row. Uses CqFieldValue.initialize() to interpret
                 * the value with respect to the type of the field and then
                 * sets the updated CqFieldValue into the proxy for writing to
                 * the repository. 
                 */
                public void setValueAt(Object aValue, int row, int col)
                {
                    if (fieldMetaProperties[col].getRoot()
                                    .equals(StpProperty.VALUE)) {
                        CqFieldValue<Object> field = 
                            (CqFieldValue<Object>)fields.get(row);
                        
                        try {
                            field.initialize(aValue);
                        } catch (WvcmException e) {
                            Utilities.exception(frame, title, e);
                        }
                        record.setFieldInfo(field.getFieldName(), field);
                    }
                }
                private static final long serialVersionUID = 1L;
            };
            final JTable table = new JTable(dataModel);

            table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
            show.setEnabled(false);
            add.setEnabled(false);
            remove.setEnabled(false);
            
            // Ask to be notified of selection changes.
            ListSelectionModel rowSM = table.getSelectionModel();
            rowSM.addListSelectionListener(new ListSelectionListener() {
                /**
                 * Enables the buttons in the dialog based on the type of
                 * field selected in the table.
                 */
                public void valueChanged(ListSelectionEvent e) {
                    if (!e.getValueIsAdjusting()){
                        int[] selected = table.getSelectedRows();
                        show.setEnabled(false);
                        add.setEnabled(false);
                        remove.setEnabled(false);
                        for (int i=0; i <selected.length; ++i)
                            if (ViewRecord.getRecordReferencedAt(fields, selected[i]) != null) {
                                show.setEnabled(true);
                            } else if (ViewRecord.isAttachmentList(fields, selected[i])) {
                                show.setEnabled(true);
                                add.setEnabled(true);
                                remove.setEnabled(true);
                            }
                    }
                }
            });

            buttons.add(show);
            show.addActionListener(new ActionListener(){
                    /**
                     * Invokes a view method on the value of the field selected
                     * in the table.
                     */
                    public void actionPerformed(ActionEvent arg0)
                    {
                        int[] selected = table.getSelectedRows();
                        
                        for (int i =0; i < selected.length; ++i) {
                            int row = selected[i];
                            CqRecord record = 
                                ViewRecord.getRecordReferencedAt(fields, row);
                            
                            if (record != null) {
                                view(record);
                            } else if (ViewRecord.isAttachmentList(fields, row)) {
                                view(ViewRecord.selectAttachment(frame, 
                                                              fields, row, 
                                                              "View"));
                            }
                        }
                    }
                });
    
            JButton deliver = new JButton("Deliver");

            buttons.add(deliver);
            deliver.addActionListener(new ActionListener(){
                    /**
                     * Delivers the modified record to the repository,
                     * closes the (now empty) change context, deletes the
                     * current viewer, and opens a new viewer on the updated
                     * record (in its original change context).
                     */
                    public void actionPerformed(ActionEvent arg0)
                    {
                        try {
                            int mode = frame.getDefaultCloseOperation();
                            
                            record.doWriteProperties(null, CqProvider.DELIVER);
                            frame.dispose();
                            view(selected).setDefaultCloseOperation(mode);
                        } catch (WvcmException ex) {
                            Utilities.exception(frame, "Deliver failed", ex);
                        }
                    }
                });
    
            buttons.add(cancel);
            cancel.addActionListener(new ActionListener(){
                    /**
                     * Discards all modifications to the record,
                     * closes the (now empty) change context, deletes the
                     * current viewer, and opens a new viewer on the original
                     * record (in its original change context).
                     */
                    public void actionPerformed(ActionEvent arg0)
                    {
                        try {
                            int mode = frame.getDefaultCloseOperation();
                            
                            record.doRevert(null);
                            frame.dispose();
                            
                            if (mode == JFrame.EXIT_ON_CLOSE)
                                System.exit(0);
                            
                            view(selected)
                                .setDefaultCloseOperation(mode);
                        } catch (WvcmException ex) {
                            Utilities.exception(frame, "Cancel failed", ex);
                        }
                    }
                });
    
            buttons.add(add);
            add.addActionListener(new ActionListener(){
                /**
                 * Adds an file specified by the user to the database as an
                 * attachment to the selected record field.
                 */
                public void actionPerformed(ActionEvent arg0)
                {
                    JFileChooser chooser = new JFileChooser();
                    int returnVal = chooser.showOpenDialog(frame);

                    if (returnVal == JFileChooser.APPROVE_OPTION) {
                        try {
                           String filename = 
                               chooser.getSelectedFile().getAbsolutePath();
                           CqFieldValue field = 
                               (CqFieldValue)fields.get(table.getSelectedRow());
                           StpLocation aLoc = (StpLocation) 
                               ((Folder)field.getValue()).location()
                                   .child("new" + System.currentTimeMillis());
                           CqAttachment attachment = 
                               ((CqProvider)record.provider()).cqAttachment(aLoc);
                           
                           attachment = attachment
                               .doCreateAttachment(filename, 
                                                          null, CqProvider.HOLD);
                           
                           JOptionPane.showMessageDialog(frame, 
                              "Added '" + filename + "' as " + attachment 
                              + " (pending delivery)");
                        } catch(Throwable t) 
                            { Utilities.exception(frame, "Add Attachment", t);}
                    }}});
            
            buttons.add(remove);
            remove.addActionListener(new ActionListener(){
                /**
                 * Removes a user-specified entry from the attachment folder 
                 * of the selected attachment field
                 */
                public void actionPerformed(ActionEvent arg0)
                {
                    try {
                       CqAttachment attachment = ViewRecord.selectAttachment
                           (frame, fields, table.getSelectedRow(), "Remove");
                       
                       if (attachment != null)
                           attachment.doUnbindAll(null);
                    } catch(Throwable t) 
                        { Utilities.exception(frame, "Remove Attachment", t);}
                    }
                }
            );
            
            panel.add(new JScrollPane(table), BorderLayout.CENTER);
            panel.add(buttons, BorderLayout.SOUTH);
            frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
            frame.setContentPane(panel);
            frame.setBounds(300, 300, 600, 300);
            frame.setVisible(true);
            
            return frame;
        }
    }
    
    /** The field meta-properties requested and displayed */
    static PropertyRequestItem.NestedPropertyName<?>[] fieldMetaProperties 
        = ViewRecord.fieldMetaProperties;
                  
    /** The record properties read prior to editing the record */
    final static PropertyRequest RECORD_PROPERTIES =
        new PropertyRequest(
            CqRecord.USER_FRIENDLY_LOCATION,
            CqRecord.STABLE_LOCATION,
            CqRecord.LEGAL_ACTIONS.nest(
                CqAction.USER_FRIENDLY_LOCATION,
                CqAction.DISPLAY_NAME,
                CqAction.TYPE),
            CqRecord.ALL_FIELD_VALUES.nest(
                StpProperty.VALUE.nest(fieldMetaProperties))
        );
    
    /**
     * Extends the ViewRecord example by adding buttons on the record view
     * window to start an edit action, add an attachment, or remove an 
     * attachment
     * @param args not used
     */
    public static void main(
        String[] args)
    {
        try {
            CqProvider provider = Utilities.getProvider().cqProvider();
            ExecuteQuery.run("Edit Record", provider, new Viewer(provider));
        } catch(Throwable ex) {
            Utilities.exception(null, "Edit Record", ex);
            System.exit(0);
        }
    }
    
    /**
     * Determines if the given PropertyName's have the same root name and
     * namespace (ignoring any nested properties they may have).
     * @param n1 A PropertyName. Must not be null
     * @param n2 Another PropertyName. Must not be null
     * @return true if n1 and n2 have the same root property name.
     */
    static boolean matches(PropertyName n1, PropertyName n2)
    {
        String ns = n1.getNamespace();
        
        return n1.getName().equals(n2.getName()) &&
            (ns == null? n2.getNamespace()==null: ns.equals(n2.getNamespace()));
    }
}
